/* Copyright (C) 2004 - 2006  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
using System;
using System.IO;
using Cecil.FlowAnalysis.CecilUtilities;
using Mono.Cecil;

namespace Cecil.FlowAnalysis.CodeStructure {
	/// <summary>
	/// </summary>
	public class ExpressionPrinter : AbstractCodeStructureVisitor {
		public static string ToString (IExpression expression)
		{
			ExpressionPrinter printer = new ExpressionPrinter ();
			expression.Accept (printer);
			return printer.Writer.ToString ();
		}

		private TextWriter _writer;

		public ExpressionPrinter ()
		{
			_writer = new StringWriter ();
		}

		public ExpressionPrinter (TextWriter writer)
		{
			if (null == writer) throw new ArgumentNullException ("writer");
			_writer = writer;
		}

		public TextWriter Writer {
			get { return _writer; }
		}

		public override void Visit (IAssignExpression node)
		{
			Visit (node.Target);
			Write (" = ");
			Visit (node.Expression);
		}

		public override void Visit (IBinaryExpression node)
		{
			Write ("(");
			Visit (node.Left);
			Write (" {0} ", ToString (node.Operator));
			Visit (node.Right);
			Write (")");
		}

		public override void Visit (IArgumentReferenceExpression node)
		{
			Write (node.Parameter.Name);
		}

		public override void Visit (IThisReferenceExpression node)
		{
			Write ("this");
		}

		public override void Visit (IFieldReferenceExpression node)
		{
			Visit (node.Target);
			Write (".{0}", node.Field.Name);
		}

		public override void Visit (IMethodReferenceExpression node)
		{
			IMethodReference method = node.Method;
			if (null == node.Target) {
				Write (CecilFormatter.FormatTypeReference (method.DeclaringType));
			} else {
				Visit (node.Target);
			}
			Write (".{0}", method.Name);
		}

		public override void Visit (IMethodInvocationExpression node)
		{
			Visit (node.Target);
			Write ("(");
			for (int i=0; i < node.Arguments.Count; ++i) {
				if (i > 0) Write (", ");
				Visit (node.Arguments [i]);
			}
			Write (")");
		}

		public override void Visit (IVariableReferenceExpression node)
		{
			string name = "local" + node.Variable.Index;
			Write (name);
		}

		public override void Visit (ILiteralExpression node)
		{
			object value = node.Value;
			if (value is string) {
				Write ("\"{0}\"", value);
				return;
			}

			Write (value == null ? "null" : CecilFormatter.ToInvariantCultureString (value).ToLower ());
		}

		public override void Visit (IUnaryExpression node)
		{
			Write ("(");
			Write (ToString (node.Operator));
			Visit (node.Operand);
			Write (")");
		}

		private string ToString (UnaryOperator op)
		{
			switch (op) {
			case UnaryOperator.Not: return "!";

			}
			throw new ArgumentException (op.ToString (), "op");
		}

		private string ToString (BinaryOperator op)
		{
			switch (op) {
			case BinaryOperator.LogicalAnd: return "&&";
			case BinaryOperator.LogicalOr: return "||";
			case BinaryOperator.Multiply: return "*";
			case BinaryOperator.ValueEquality: return "==";
			case BinaryOperator.ValueInequality: return "!=";
			case BinaryOperator.LessThan: return "<";
			case BinaryOperator.LessThanOrEqual: return "<=";
			case BinaryOperator.GreaterThan: return ">";
			case BinaryOperator.GreaterThanOrEqual: return ">=";
			}
			throw new ArgumentException (op.ToString (), "op");
		}

		private void Write (string text)
		{
			_writer.Write (text);
		}

		private void Write (string format, params object[] args)
		{
			_writer.Write (format, args);
		}
	}
}
